(**
Copyright (C) Boldo, Lauter, Muller 2019

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*)

From Flocq Require Import Core Plus_error Mult_error Sterbenz Operations Pff2Flocq.
Open Scope R_scope.



Definition sign := fun x => 
    if (Rlt_bool 0 x) then 1 else
    if (Req_bool 0 x) then 0 else (-1).

Lemma sign_pos : forall x, 0 < x -> sign x = 1.
Proof.
intros t Ht; unfold sign; rewrite Rlt_bool_true; easy.
Qed.

Lemma sign_0 : sign 0 = 0.
Proof.
unfold sign; rewrite Rlt_bool_false; try auto with real.
rewrite Req_bool_true; easy.
Qed.

Lemma sign_neg : forall x, x < 0 -> sign x = -1.
Proof.
intros t Ht.
unfold sign; rewrite Rlt_bool_false; try auto with real.
rewrite Req_bool_false; try auto with real.
Qed.


Lemma round_N_small : forall (beta : radix) (fexp : Z -> Z)
         (choice : Z -> bool) (x : R) (ex : Z),
       bpow beta (ex - 1) <= Rabs x < bpow beta ex ->
       (ex < fexp ex)%Z ->
       round beta fexp (Znearest choice) x = 0.
Proof with auto with typeclass_instances.
intros beta fexp choice x ex H1 H2.
case (Rle_or_lt 0 x); intros Hx.
apply round_N_small_pos with ex; try easy.
rewrite <- (Rabs_right x); try easy; auto with real.
rewrite <- (Ropp_involutive x), <- Ropp_0.
rewrite round_N_opp; f_equal.
apply round_N_small_pos with ex; try easy.
rewrite <- Rabs_left; easy.
Qed.





Lemma round_N_ge_ge_midp : 
       forall (beta : radix) (fexp : Z -> Z),
       Valid_exp fexp ->
       forall (choice : Z -> bool) (u v : R),
       generic_format beta fexp u ->
       u <= round beta fexp (Znearest choice) v ->
       (u + pred beta fexp u) / 2 <= v.
Proof with auto with typeclass_instances.
intros beta fexp H1 choice u v Hu H2.
assert (K: (u=0 /\ negligible_exp fexp = None) \/ (pred beta fexp u < u)).
case (Req_dec u 0); intros Zu.
case_eq (negligible_exp fexp).
intros n Hn; right.
rewrite Zu, pred_0.
unfold ulp; rewrite Req_bool_true, Hn; try easy.
rewrite <- Ropp_0.
apply Ropp_lt_contravar, bpow_gt_0.
intros _; left; split; easy.
right.
apply pred_lt_id...
(* *)
case K.
intros (K1,K2).
(* . *)
rewrite K1, pred_0.
unfold ulp; rewrite Req_bool_true, K2; try easy.
replace ((0+-0)/2) with 0 by field.
case (Rle_or_lt 0 v); try easy.
intros H3; contradict H2.
rewrite K1; apply Rlt_not_le.
assert (H4: round beta fexp (Znearest choice) v <= 0).
apply round_le_generic...
apply generic_format_0...
now left.
case H4; try easy.
intros H5.
absurd (v=0); try auto with real.
apply eq_0_round_0_negligible_exp with beta fexp (Znearest choice)...
(* . *)
intros K1.
case (Rle_or_lt ((u + pred beta fexp u) / 2) v); try easy.
intros H3.
absurd (u <= round beta fexp (Znearest choice) v); try easy.
apply Rlt_not_le.
apply Rle_lt_trans with (2:=K1).
apply round_N_le_midp...
apply generic_format_pred...
rewrite succ_pred...
apply Rlt_le_trans with (1:=H3).
right; f_equal; ring.
Qed.


Lemma round_N_le_le_midp : 
       forall (beta : radix) (fexp : Z -> Z),
       Valid_exp fexp ->
       forall (choice : Z -> bool) (u v : R),
       generic_format beta fexp u ->
       round beta fexp (Znearest choice) v <= u ->
       v <= (u + succ beta fexp u) / 2.
Proof with auto with typeclass_instances.
intros beta fexp H1 choice u v Hu H2.
apply Ropp_le_cancel.
apply Rle_trans with (((-u)+pred beta fexp (-u))/2).
rewrite pred_opp; right; field.
apply round_N_ge_ge_midp with
   (choice := fun t:Z => negb (choice (- (t + 1))%Z))...
apply generic_format_opp...
rewrite <- (Ropp_involutive (round _ _ _ _)).
rewrite <- round_N_opp, Ropp_involutive.
apply Ropp_le_contravar; easy.
Qed.


Lemma generic_format_FLT_mult_bpow : forall beta emin prec e f,
   Prec_gt_0 prec ->
   (0 <= e)%Z -> 
   generic_format beta (FLT_exp emin prec) f ->
   generic_format beta (FLT_exp emin prec) (f*bpow beta e).
Proof with auto with typeclass_instances.
intros beta emin prec e f Hp He Hf.
destruct (FLT_format_generic _ _ _ _ Hf).
apply generic_format_FLT.
exists (Float beta (Fnum f0) (Fexp f0+e)).
rewrite H; unfold F2R; simpl.
rewrite bpow_plus; ring.
simpl; easy.
simpl; omega.
Qed.

Lemma generic_format_FLT_other: forall beta emin prec e f,
   Prec_gt_0 prec ->
   (exists n:Z, f = IZR n*bpow beta e) ->
   (emin <= e)%Z ->
   (Rabs f < bpow beta (prec+e)) ->
   generic_format beta (FLT_exp emin prec) f.
Proof with auto with typeclass_instances.
intros beta emin prec e f H1 (n,Hn) H2 H3.
apply generic_format_FLT.
exists (Float beta n e).
unfold F2R; simpl; easy.
2: simpl; easy.
simpl; apply lt_IZR.
rewrite IZR_Zpower, abs_IZR.
2: unfold Prec_gt_0 in H1; omega.
apply Rmult_lt_reg_r with (bpow beta e).
apply bpow_gt_0.
rewrite <- bpow_plus.
apply Rle_lt_trans with (2:=H3).
right; rewrite Hn, Rabs_mult.
rewrite (Rabs_right (bpow _ _)); try easy.
apply Rle_ge, bpow_ge_0.
Qed.

Lemma generic_format_FLT_other': forall beta emin prec e x (f:float beta),
   Prec_gt_0 prec ->
   (x= F2R f) ->
   (emin <= e <= Fexp f)%Z ->
   (Rabs x < bpow beta (prec+e)) ->
   generic_format beta (FLT_exp emin prec) x.
Proof with auto with typeclass_instances.
intros beta emin prec e x f H1 H2 (H3,H4) H5.
apply generic_format_FLT_other with e; try easy.
exists (Fnum f*beta^(Fexp f-e))%Z.
rewrite H2; unfold F2R.
rewrite mult_IZR, Rmult_assoc; f_equal.
rewrite IZR_Zpower; try omega.
rewrite <- bpow_plus.
f_equal; ring.
Qed.



Notation Znearest0 := (Znearest (fun x => (Zlt_bool x 0))).

Section About_rnd_NO.

Definition Rnd_N0_pt (F : R -> Prop) (x f : R) :=
  Rnd_N_pt F x f /\
  forall f2 : R, Rnd_N_pt F x f2 -> (Rabs f <= Rabs f2)%R.

Definition Rnd_N0 (F : R -> Prop) (rnd : R -> R) :=
  forall x : R, Rnd_N0_pt F x (rnd x).

Theorem Rnd_N0_NG_pt :
  forall F : R -> Prop,
  F 0 ->
  forall x f,
  Rnd_N0_pt F x f <-> Rnd_NG_pt F (fun x f => Rabs f <= Rabs x) x f.
Proof.
intros F HF x f.
destruct (Rle_or_lt 0 x) as [Hx|Hx].
(* *)
split ; intros (H1, H2).
(* . *)
assert (Hf := Rnd_N_pt_ge_0 F HF x f Hx H1).
split.
exact H1.
destruct (Rnd_N_pt_DN_or_UP _ _ _ H1) as [H3|H3].
(* . . *)
left.
rewrite Rabs_pos_eq with (1 := Hf).
rewrite Rabs_pos_eq with (1 := Hx).
apply H3.
(* . . *)
right.
intros f2 Hxf2.
specialize (H2 _ Hxf2).
destruct (Rnd_N_pt_DN_or_UP _ _ _ Hxf2) as [H4|H4].
apply Rle_antisym.
apply Rle_trans with x.
apply H4.
apply H3.
rewrite Rabs_pos_eq with (1 := Hf) in H2.
rewrite Rabs_pos_eq in H2.
exact H2.
now apply Rnd_N_pt_ge_0 with F x.
eapply Rnd_UP_pt_unique ; eassumption.
(* . *)
split.
exact H1.
intros f2 Hxf2.
destruct H2 as [H2|H2].
assert (Hf := Rnd_N_pt_ge_0 F HF x f Hx H1).
assert (Hf2 := Rnd_N_pt_ge_0 F HF x f2 Hx Hxf2).
rewrite 2!Rabs_pos_eq ; trivial.
rewrite 2!Rabs_pos_eq in H2 ; trivial.
destruct (Rnd_N_pt_DN_or_UP _ _ _ Hxf2) as [H3|H3].
apply H3.
apply H1.
apply H2.
apply Rle_trans with (1 := H2).
apply H3.
rewrite (H2 _ Hxf2).
apply Rle_refl.
(* *)
assert (Hx' := Rlt_le _ _ Hx).
clear Hx. rename Hx' into Hx.
split ; intros (H1, H2).
(* . *)
assert (Hf := Rnd_N_pt_le_0 F HF x f Hx H1).
split.
exact H1.
destruct (Rnd_N_pt_DN_or_UP _ _ _ H1) as [H3|H3].
(* . . *)
right.
intros f2 Hxf2.
specialize (H2 _ Hxf2).
destruct (Rnd_N_pt_DN_or_UP _ _ _ Hxf2) as [H4|H4].
eapply Rnd_DN_pt_unique ; eassumption.
apply Rle_antisym.
2: apply Rle_trans with x.
2: apply H3.
2: apply H4.
rewrite Rabs_left1 with (1 := Hf) in H2.
rewrite Rabs_left1 in H2.
now apply Ropp_le_cancel.
now apply Rnd_N_pt_le_0 with F x.
(* . . *)
left.
rewrite Rabs_left1 with (1 := Hf).
rewrite Rabs_left1 with (1 := Hx).
apply Ropp_le_contravar.
apply H3.
(* . *)
split.
exact H1.
intros f2 Hxf2.
destruct H2 as [H2|H2].
assert (Hf := Rnd_N_pt_le_0 F HF x f Hx H1).
assert (Hf2 := Rnd_N_pt_le_0 F HF x f2 Hx Hxf2).
rewrite 2!Rabs_left1 ; trivial.
rewrite 2!Rabs_left1 in H2 ; trivial.
apply Ropp_le_contravar.
apply Ropp_le_cancel in H2.
destruct (Rnd_N_pt_DN_or_UP _ _ _ Hxf2) as [H3|H3].
2: apply H3.
2: apply H1.
2: apply H2.
apply Rle_trans with (2 := H2).
apply H3.
rewrite (H2 _ Hxf2).
apply Rle_refl.
Qed.

Lemma Rnd_N0_pt_unique_prop :
  forall F : R -> Prop,
  F 0 ->
  Rnd_NG_pt_unique_prop F (fun x f => Rabs f <= Rabs x).
Proof.
intros F HF x d u Hxd1 Hxd2 Hxu1 Hxu2 Hd Hu.
apply Rle_antisym.
apply Rle_trans with x.
apply Hxd1.
apply Hxu1.
destruct (Rle_or_lt 0 x) as [Hx|Hx].
apply Hxd1.
apply Hxu1.
rewrite Rabs_pos_eq with (1 := Hx) in Hu.
rewrite Rabs_pos_eq in Hu.
exact Hu.
apply Rle_trans with (1:=Hx).
apply Hxu1.
(* *)
apply Hxu1.
apply Hxd1.
rewrite Rabs_left with (1 := Hx) in Hd.
rewrite Rabs_left1 in Hd.
now apply Ropp_le_cancel.
apply Rlt_le, Rle_lt_trans with (2:=Hx).
apply Hxd1.
Qed.

Theorem Rnd_N0_pt_unique :
  forall F : R -> Prop,
  F 0 ->
  forall x f1 f2 : R,
  Rnd_N0_pt F x f1 -> Rnd_N0_pt F x f2 ->
  f1 = f2.
Proof.
intros F HF x f1 f2 H1 H2.
apply (Rnd_NG_pt_unique F _ (Rnd_N0_pt_unique_prop F HF) x).
now apply -> Rnd_N0_NG_pt.
now apply -> Rnd_N0_NG_pt.
Qed.

Theorem Rnd_N0_pt_N :
  forall F : R -> Prop,
  F 0 ->
  forall x f : R,
  Rnd_N_pt F x f ->
  (Rabs f <= Rabs x)%R ->
  Rnd_N0_pt F x f.
Proof.
intros F HF x f Rxf Hxf.
split.
apply Rxf.
intros g Rxg.
destruct (Rabs_eq_Rabs (f - x) (g - x)) as [H|H].
apply Rle_antisym.
apply Rxf.
apply Rxg.
apply Rxg.
apply Rxf.
(* *)
replace g with f.
apply Rle_refl.
apply Rplus_eq_reg_r with (1 := H).
(* *)
assert (g = 2 * x - f)%R.
replace (2 * x - f)%R with (x - (f - x))%R by ring.
rewrite H.
ring.
destruct (Rle_lt_dec 0 x) as [Hx|Hx].
(* . *)
revert Hxf.
rewrite Rabs_pos_eq with (1 := Hx).
rewrite 2!Rabs_pos_eq ; try ( apply (Rnd_N_pt_ge_0 F HF x) ; assumption ).
intros Hxf.
rewrite H0.
apply Rplus_le_reg_r with f.
ring_simplify.
apply Rmult_le_compat_l with (2 := Hxf).
now apply IZR_le.
(* . *)
revert Hxf.
apply Rlt_le in Hx.
rewrite Rabs_left1 with (1 := Hx).
rewrite 2!Rabs_left1 ; try ( apply (Rnd_N_pt_le_0 F HF x) ; assumption ).
intros Hxf.
rewrite H0.
apply Ropp_le_contravar.
apply Rplus_le_reg_r with f.
ring_simplify.
apply Rmult_le_compat_l.
now apply IZR_le.
now apply Ropp_le_cancel.
Qed.

Theorem Rnd_N0_unique :
  forall (F : R -> Prop),
  F 0 ->
  forall rnd1 rnd2 : R -> R,
  Rnd_N0 F rnd1 -> Rnd_N0 F rnd2 ->
  forall x, rnd1 x = rnd2 x.
Proof.
intros F HF rnd1 rnd2 H1 H2 x.
now apply Rnd_N0_pt_unique with F x.
Qed.

Theorem Rnd_N0_pt_monotone :
  forall F : R -> Prop,
  F 0 ->
  round_pred_monotone (Rnd_N0_pt F).
Proof.
intros F HF x y f g Hxf Hyg Hxy.
apply (Rnd_NG_pt_monotone F _ (Rnd_N0_pt_unique_prop F HF) x y).
now apply -> Rnd_N0_NG_pt.
now apply -> Rnd_N0_NG_pt.
exact Hxy.
Qed.

Theorem Rnd_N0_pt_refl :
  forall F : R -> Prop,
  forall x : R, F x ->
  Rnd_N0_pt F x x.
Proof.
intros F x Hx.
split.
now apply Rnd_N_pt_refl.
intros f Hxf.
apply Req_le.
apply f_equal.
now apply sym_eq, Rnd_N_pt_idempotent with (1 := Hxf).
Qed.

Theorem Rnd_N0_pt_idempotent :
  forall F : R -> Prop,
  forall x f : R,
  Rnd_N0_pt F x f -> F x ->
  f = x.
Proof.
intros F x f (Hf,_) Hx.
now apply Rnd_N_pt_idempotent with F.
Qed.

Theorem satisfies_any_imp_N0 :
  forall F : R -> Prop,
  F 0 -> satisfies_any F ->
  round_pred (Rnd_N0_pt F).
Proof.
intros F HF0 Hany.
split.
assert (H : round_pred_total (Rnd_NG_pt F (fun a b => (Rabs b <= Rabs a)%R))).
apply satisfies_any_imp_NG.
apply Hany.
intros x d u Hf Hd Hu.
destruct (Rle_lt_dec 0 x) as [Hx|Hx].
right.
rewrite Rabs_pos_eq with (1 := Hx).
rewrite Rabs_pos_eq.
apply Hd.
apply Hd; try easy.
left.
rewrite Rabs_left with (1 := Hx).
rewrite Rabs_left1.
apply Ropp_le_contravar.
apply Hu.
apply Hu; try easy.
now apply Rlt_le.
intros x.
destruct (H x) as (f, Hf).
exists f.
apply <- Rnd_N0_NG_pt.
apply Hf.
apply HF0.
apply Rnd_N0_pt_monotone.
apply HF0.
Qed.



Variable beta : radix.
Variable fexp : Z -> Z.
Context { valid_exp : Valid_exp fexp }.


Global Instance valid_rnd_N0 : Valid_rnd Znearest0 := valid_rnd_N _.

Theorem round_N0_pt :
  forall x,
  Rnd_N0_pt (generic_format beta fexp) x (round beta fexp Znearest0 x).
Proof.
intros x.
generalize (round_N_pt beta fexp (fun t => Zlt_bool t 0) x).
set (f := round beta fexp (Znearest (fun t => Zlt_bool t 0)) x).
intros Rxf.
destruct (Req_dec (x - round beta fexp Zfloor x) (round beta fexp Zceil x - x)) as [Hm|Hm].
(* *)
apply Rnd_N0_pt_N.
apply generic_format_0.
exact Rxf.
destruct (Rle_or_lt 0 x) as [Hx|Hx].
(* . *)
rewrite Rabs_pos_eq with (1 := Hx).
rewrite Rabs_pos_eq.
unfold f.
rewrite round_N_middle with (1 := Hm).
rewrite Zlt_bool_false.
now apply round_DN_pt.
apply Zfloor_lub.
apply Rmult_le_pos with (1 := Hx).
apply bpow_ge_0.
apply Rnd_N_pt_ge_0 with (2 := Hx) (3 := Rxf).
apply generic_format_0.
(* . *)
rewrite Rabs_left with (1 := Hx).
rewrite Rabs_left1.
apply Ropp_le_contravar.
unfold f.
rewrite round_N_middle with (1 := Hm).
rewrite Zlt_bool_true.
now apply round_UP_pt.
apply lt_IZR.
apply Rle_lt_trans with (scaled_mantissa beta fexp x).
apply Zfloor_lb.
simpl.
rewrite <- (Rmult_0_l (bpow beta (- (cexp beta fexp x)))).
apply Rmult_lt_compat_r with (2 := Hx).
apply bpow_gt_0.
apply Rnd_N_pt_le_0 with (3 := Rxf).
apply generic_format_0.
now apply Rlt_le.
(* *)
split.
apply Rxf.
intros g Rxg.
rewrite Rnd_N_pt_unique with (3 := Hm) (4 := Rxf) (5 := Rxg).
apply Rle_refl.
apply round_DN_pt; easy.
apply round_UP_pt; easy.
Qed.


Lemma round_N0_opp :
  forall x,
  (round beta fexp Znearest0 (- x) = - round beta fexp Znearest0 x)%R.
Proof.
intros x.
rewrite round_N_opp.
apply Ropp_eq_compat.
apply round_ext.
clear x; intro x.
unfold Znearest.
case_eq (Rcompare (x - IZR (Zfloor x)) (/ 2)); intro C;
[|reflexivity|reflexivity].
apply Rcompare_Eq_inv in C.
assert (H : negb (- (Zfloor x + 1) <? 0)%Z = (Zfloor x <? 0)%Z);
  [|now rewrite H].
rewrite negb_Zlt_bool.
case_eq (Zfloor x <? 0)%Z; intro C'.
apply Zlt_is_lt_bool in C'.
apply Zle_bool_true.
omega.
apply Z.ltb_ge in C'.
apply Zle_bool_false.
omega.
Qed.

Lemma round_N_0_eq_N0: forall c x,
  round beta fexp (Znearest c) x = 0 ->
  round beta fexp Znearest0 x = 0.
Proof with auto with typeclass_instances.
intros c x H1.
rewrite <- H1.
apply Rnd_N0_pt_unique with (generic_format beta fexp) x.
apply generic_format_0...
apply round_N0_pt...
apply Rnd_N0_pt_N...
apply generic_format_0...
apply round_N_pt...
rewrite H1, Rabs_R0.
apply Rabs_pos.
Qed.



Lemma round_N_eq_ties: forall c1 c2 x, 
   x - round beta fexp Zfloor x <> round beta fexp Zceil x - x ->
  round beta fexp (Znearest c1) x = round beta fexp (Znearest c2) x.
Proof with auto with typeclass_instances.
intros c1 c2 x.
pose (d:=round beta fexp Zfloor x); pose (u:=round beta fexp Zceil x); fold d; fold u; intros H.
case (Rle_or_lt ((d+u)/2) x); intros L.
2:rewrite 2!round_N_eq_DN...
destruct L as [L|L].
rewrite 2!round_N_eq_UP...
contradict H; rewrite <- L; field.
Qed.


End About_rnd_NO.






